// ----------------------------------------------------------------------------
// Project:	Implementation Alternatives for Bottom-Up Evaluation
// Filename:	test5push.cpp
// Purpose:	Time Measurements for Rule Implementations
// Last Change:	06.02.2010
// Language:	C++
// EMail:	brass@acm.org
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Univ. Halle, Von-Seckendorff-Platz 1, D-06120 Halle, GERMANY
// Copyright:	(c) 2010 by Stefan Brass
//		Downloads and modifications for tests and scientific purposes
//		permitted, provided that some acknowledgement of my work
//		remains, and modified versions are clearly marked as such.
//		There is no warranty at all - this code may contain bugs.
// ----------------------------------------------------------------------------

// Example Program 5:
//
// p1(A) :- q(A).
// p2(A,1) :- p1(A).
// p2(A,2) :- p1(A).
// p3(A,B,1) :- p2(A,B).
// p3(A,B,2) :- p2(A,B).
// p4(A,B,C,1) :- p3(A,B,C).
// p4(A,B,C,2) :- p3(A,B,C).
// p5(A,B,C,D,1) :- p4(A,B,C,D).
// p5(A,B,C,D,2) :- p4(A,B,C,D).
// p6(A,B,C,D,E,1) :- p5(A,B,C,D,E).
// p6(A,B,C,D,E,2) :- p5(A,B,C,D,E).
// p7(A,B,C,D,E,F,1) :- p6(A,B,C,D,E,F).
// p7(A,B,C,D,E,F,2) :- p6(A,B,C,D,E,F).

// q := {(i)|i = 1,...,50000}


#include <iostream>
using namespace std;

#if defined(_MSC_VER) || defined(_WINDOWS_)
#define WINDOWS
#else
#define UNIX
#endif

// ----------------------------------------------------------------------------
// Sizes for test:
// ----------------------------------------------------------------------------

const int Q_TUPLES = 50000;

const bool PRINT_EXTENSION = false;

const int STACKSIZE = 50;

// ----------------------------------------------------------------------------
// Timing Functions (UNIX Version):
// ----------------------------------------------------------------------------

#ifdef UNIX
#include <sys/time.h>
#include <time.h>

class timer {
public:
	void start()
	{
		gettimeofday(&Time_Start, 0);
		Clock_Start = clock();
	}

	void end()
	{
		gettimeofday(&Time_End, 0);
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		long sec = Time_End.tv_sec - Time_Start.tv_sec;
		long usec = Time_End.tv_usec - Time_Start.tv_usec;
		if(usec < 0) {
			usec += 1000000;
			sec -= 1;
		}
		return sec * 1000 + (usec + 500) / 1000;
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	timeval Time_Start, Time_End;
	clock_t Clock_Start, Clock_End;
};
#endif

// ----------------------------------------------------------------------------
// Timing Functions (Windows Version):
// ----------------------------------------------------------------------------

#ifdef WINDOWS
#include <windows.h>
#include <time.h>

class timer {
public:
	void start()
	{
		GetSystemTimeAsFileTime(&(Start.ft));
		Clock_Start = clock();
	}

	void end()
	{
		GetSystemTimeAsFileTime(&(End.ft));
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		return static_cast<long> ((End.ns100 - Start.ns100) / 10000);
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	union {
		long long ns100;
		FILETIME ft;
	} Start, End;
	clock_t Clock_Start, Clock_End;
};
#endif


// ----------------------------------------------------------------------------
// Stack (of int):
// ----------------------------------------------------------------------------

class stack {

private:
	static int stack_arr[STACKSIZE];
	static int *stack_p;
	static int stack_free;

public:
	static inline void stack_init()
	{
		stack_p = stack_arr;
		stack_free = STACKSIZE;
	}

	static inline void push(int i)
	{
		if(stack_free > 0) {
			stack_free--;
			*(stack_p++) = i;
		}
		else {
			cout << "Stack overflow\n";
			exit(1);
		}
	}

	static inline int pop()
	{
		if(stack_p > stack_arr) {
			stack_free++;
			return *(--stack_p);
		}
		else {
			cout << "Stack underflow\n";
			exit(1);
		}
	}

	static inline bool is_empty()
	{
		return (stack_p == stack_arr);
	}
};

int stack::stack_arr[STACKSIZE];
int *stack::stack_p;
int stack::stack_free;


//-----------------------------------------------------------------------------
// Predicate q: Tuples generated by computation
//-----------------------------------------------------------------------------

class q_cursor {
public:
	q_cursor() { }

	void open() { i = 0; }

	bool fetch()
	{
		if(i < Q_TUPLES) {
			i++;
			return true;
		}
		else
			return false;
	}

	int col_1() const { return i; }

	void close() {}

	void push() { stack::push(i); }

	void pop() { i = stack::pop(); }


private:
	int i;
};


// ----------------------------------------------------------------------------
// Predicate p7 - Push Method
// ----------------------------------------------------------------------------

class p7_cursor {
private:

	// Continuation labels:
	typedef enum {CONT_init, CONT_q, CONT_p1, CONT_p2, CONT_p3, CONT_p4,
       			CONT_p5, CONT_p6 } cont_t;

	// Functions for using the global stack:
	inline void push_int(int i)
	{
		stack::push(i);
	}

	inline int pop_int()
	{
		return stack::pop();
	}

	inline void push_cont(cont_t t)
	{
		stack::push(t);
	}

	inline cont_t pop_cont()
	{
		return (cont_t) stack::pop();
	}

public:
	void open()
	{
		stack::stack_init();
		push_cont(CONT_init);
	}

	bool fetch()
	{
		// Arguments:
		int p1_1;
		int p2_1, p2_2;
		int p3_1, p3_2, p3_3;
		int p4_1, p4_2, p4_3, p4_4;
		int p5_1, p5_2, p5_3, p5_4, p5_5;
		int p6_1, p6_2, p6_3, p6_4, p6_5, p6_6;

		// Cursors:
		q_cursor q_c;

		// Fetch task from stack:
		if(stack::is_empty())
			return false;
		switch(pop_cont()) {
			case CONT_init: goto cont_init;
			case CONT_q: goto cont_q;
			case CONT_p1: goto cont_p1;
			case CONT_p2: goto cont_p2;
			case CONT_p3: goto cont_p3;
			case CONT_p4: goto cont_p4;
			case CONT_p5: goto cont_p5;
			case CONT_p6: goto cont_p6;
			default: cout << "Impossible cont\n"; return false;
		}

	cont_init:
		// p1(A) :- q(A).
		q_c.open();
		if(!q_c.fetch())
			return false;
		p1_1 = q_c.col_1();
		q_c.push();
		push_cont(CONT_q);
		goto p1;
	
	cont_q:
		// p1(A) :- q(A).
		q_c.pop();
		if(!q_c.fetch())
			return false;
		p1_1 = q_c.col_1();
		q_c.push();
		push_cont(CONT_q);
		goto p1;

	p1:
		// p2(A,1) :- p1(A).
		p2_1 = p1_1;
		p2_2 = 1;
		push_int(p1_1);
		push_cont(CONT_p1);
		goto p2;

	cont_p1:
		// p2(A,2) :- p1(A).
		p1_1 = pop_int();
		p2_1 = p1_1;
		p2_2 = 2;
		goto p2;

	p2:
		// p3(A,B,1) :- p2(A,B).
		p3_1 = p2_1;
		p3_2 = p2_2;
		p3_3 = 1;
		push_int(p2_1);
		push_int(p2_2);
		push_cont(CONT_p2);
		goto p3;

	cont_p2:
		// p3(A,B,2) :- p2(A,B).
		p2_2 = pop_int();
		p2_1 = pop_int();
		p3_1 = p2_1;
		p3_2 = p2_2;
		p3_3 = 2;
		goto p3;

	p3:
		// p4(A,B,C,1) :- p3(A,B,C).
		p4_1 = p3_1;
		p4_2 = p3_2;
		p4_3 = p3_3;
		p4_4 = 1;
		push_int(p3_1);
		push_int(p3_2);
		push_int(p3_3);
		push_cont(CONT_p3);
		goto p4;

	cont_p3:
		// p4(A,B,C,2) :- p3(A,B,C).
		p3_3 = pop_int();
		p3_2 = pop_int();
		p3_1 = pop_int();
		p4_1 = p3_1;
		p4_2 = p3_2;
		p4_3 = p3_3;
		p4_4 = 2;
		goto p4;

	p4:
		// p5(A,B,C,D,1) :- p4(A,B,C,D).
		p5_1 = p4_1;
		p5_2 = p4_2;
		p5_3 = p4_3;
		p5_4 = p4_4;
		p5_5 = 1;
		push_int(p4_1);
		push_int(p4_2);
		push_int(p4_3);
		push_int(p4_4);
		push_cont(CONT_p4);
		goto p5;

	cont_p4:
		// p5(A,B,C,D,2) :- p4(A,B,C,D).
		p4_4 = pop_int();
		p4_3 = pop_int();
		p4_2 = pop_int();
		p4_1 = pop_int();
		p5_1 = p4_1;
		p5_2 = p4_2;
		p5_3 = p4_3;
		p5_4 = p4_4;
		p5_5 = 2;
		goto p5;

	p5:
		// p6(A,B,C,D,E,1) :- p5(A,B,C,D,E).
		p6_1 = p5_1;
		p6_2 = p5_2;
		p6_3 = p5_3;
		p6_4 = p5_4;
		p6_5 = p5_5;
		p6_6 = 1;
		push_int(p5_1);
		push_int(p5_2);
		push_int(p5_3);
		push_int(p5_4);
		push_int(p5_5);
		push_cont(CONT_p5);
		goto p6;

	cont_p5:
		// p6(A,B,C,D,E,2) :- p5(A,B,C,D,E).
		p5_5 = pop_int();
		p5_4 = pop_int();
		p5_3 = pop_int();
		p5_2 = pop_int();
		p5_1 = pop_int();
		p6_1 = p5_1;
		p6_2 = p5_2;
		p6_3 = p5_3;
		p6_4 = p5_4;
		p6_5 = p5_5;
		p6_6 = 2;
		goto p6;

	p6:
		// p7(A,B,C,D,E,F,1) :- p6(A,B,C,D,E,F).
		p7_1 = p6_1;
		p7_2 = p6_2;
		p7_3 = p6_3;
		p7_4 = p6_4;
		p7_5 = p6_5;
		p7_6 = p6_6;
		p7_7 = 1;
		push_int(p6_1);
		push_int(p6_2);
		push_int(p6_3);
		push_int(p6_4);
		push_int(p6_5);
		push_int(p6_6);
		push_cont(CONT_p6);
		goto p7;

	cont_p6:
		// p7(A,B,C,D,E,F,2) :- p6(A,B,C,D,E,F).
		p6_6 = pop_int();
		p6_5 = pop_int();
		p6_4 = pop_int();
		p6_3 = pop_int();
		p6_2 = pop_int();
		p6_1 = pop_int();
		p7_1 = p6_1;
		p7_2 = p6_2;
		p7_3 = p6_3;
		p7_4 = p6_4;
		p7_5 = p6_5;
		p7_6 = p6_6;
		p7_7 = 2;
		goto p7;

	p7:
		return true;
	}

    int col_1() const { return p7_1; }
    int col_2() const { return p7_2; }
    int col_3() const { return p7_3; }
    int col_4() const { return p7_4; }
    int col_5() const { return p7_5; }
    int col_6() const { return p7_6; }
    int col_7() const { return p7_7; }

    void close()
    {
    }
private:
	// Arguments:
	int p7_1, p7_2, p7_3, p7_4, p7_5, p7_6, p7_7;
};

//-----------------------------------------------------------------------------
// Main Program:
//-----------------------------------------------------------------------------

int main()
{
	timer t;
	int c1_must_be = 1;
	int num_repetitions = 0;
	int c1 = 0;
	int c2 = 0;
	int c3 = 0;
	int c4 = 0;
	int c5 = 0;
	int c6 = 0;
	int c7 = 0;

	cout << "Test of Method 3: Push [Program 5]\n";
	t.start();
	p7_cursor c;
	c.open();
	int num_tuples = 0;
	while(c.fetch()) {
		c1 = c.col_1();
		if(c1 != c1_must_be) {
			cout << "Error: Wrong value for c1:" << c.col_1()
				<< ", must be: " << c1_must_be << "!\n";
			exit(1);
		}
		num_repetitions++;
		if(num_repetitions == 64) {
			c1_must_be++;
			num_repetitions = 0;
		}
		c2 += c.col_2();
		c3 += c.col_3();
		c4 += c.col_4();
		c5 += c.col_5();
		c6 += c.col_6();
		c7 += c.col_7();
		num_tuples++;
	}
	c.close();
	t.end();

	if(PRINT_EXTENSION) {
		cout << "Extension of p6:\n\n";
		c.open();
		while(c.fetch()) {
			cout << "p7(" << c.col_1() << ", " << c.col_2()
				<< ", " << c.col_3() << ", " << c.col_4()
				<< ", " << c.col_5() << ", " << c.col_6()
				<< ", " << c.col_7() << ").\n";
		}
		c.close();
		cout << "\n";
	}

	if(c2 != Q_TUPLES * 96 )
		cout << "Error: Wrong column sum c2: " << c2 << "\n";
	if(c3 != c2 || c4 != c2 || c5 != c2 || c6 != c2)
		cout << "Error: Wrong column sum c3, c4, c5, c6!\n";
	cout << "Total number of tuples: " << num_tuples << "\n";
	if(num_tuples != Q_TUPLES * 64)
		cout << "Error: Wrong number of tuples!\n";
	cout << "\nUsed time: " << t.realtime_msec() << " ms,  "
		<< "CPU: " << t.cputime_msec() << " ms.\n\n\n";

	return 0;
}


